#! /usr/bin/env python

#############################################################################
##                                                                         ##
## rtgraph3d.py ---  Real Time 3D graph visaulizer                         ##
##              see http://www.secdev.org/projects/rtgraph3d               ##
##              for more informations                                      ##
##                                                                         ##
## Copyright (C) 2007  Philippe Biondi <phil@secdev.org>                   ##
##                                                                         ##
## This program is free software; you can redistribute it and/or modify it ##
## under the terms of the GNU General Public License version 2 as          ##
## published by the Free Software Foundation.                              ##
##                                                                         ##
## This program is distributed in the hope that it will be useful, but     ##
## WITHOUT ANY WARRANTY; without even the implied warranty of              ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        ##
## General Public License for more details.                                ##
##                                                                         ##
#############################################################################


from visual import *
from select import select
import os,sys,time,traceback,string,random
import logging,getopt,socket
import thread
import povexport
import struct
from pysqlite2 import dbapi2 as sqlite


log = logging.getLogger("rtgraph3d")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("%(levelname)-5s: %(message)s"))
log.addHandler(console_handler)
log.setLevel(20)


DT = 0.1 # time interval
ACCEL = 5


class python_physics_engine:
    FRICTION = -0.5
    MIN_DIST = 16
    
    @staticmethod
    def add_edge(e1,e2):
        pass
    @staticmethod
    def update(the_world, pos, spd, changed, dt, ATTRACT, REPULS):
        ln = len(pos)
        for o in the_world.lst:
            o.F = python_physics_engine.FRICTION*o.V
        i = 0
        while i < the_world.world_len:
            o = the_world.lst[i]
            i += 1
            j = i
            while j < the_world.world_len:
                n = the_world.lst[j]
                j += 1
                d = n.pos-o.pos
                if n in o.edges:
                    F = d*ATTRACT
                    o.F += F
                    n.F -= F
                    
                l = d.mag2
                l = l**2
                if l < python_physics_engine.MIN_DIST:
                    l = python_physics_engine.MIN_DIST
                F = -REPULS*d/l
                o.F += F
                n.F -= F
        for o in the_world.lst:
            dV = o.F*dt
            o.V += dV
            o.pos += (dV/2+o.V)*dt
        


try:
    import PyInline
    physics_engine = PyInline.build(language="C", code=r"""

#define MAXPTS 2000
#define MIN_DIST 16
#define MAX_DIST 10
#define FRICTION -0.5
#define ACCEL 1

#define NEEDED_PTS (MAXPTS+4)

float coordx[NEEDED_PTS];
float coordy[NEEDED_PTS];
float coordz[NEEDED_PTS];
float speedx[NEEDED_PTS];
float speedy[NEEDED_PTS];
float speedz[NEEDED_PTS];
float forcex[NEEDED_PTS];
float forcey[NEEDED_PTS];
float forcez[NEEDED_PTS];

PyFloatObject *pf_coordx[MAXPTS];
PyFloatObject *pf_coordy[MAXPTS];
PyFloatObject *pf_coordz[MAXPTS];
PyFloatObject *pf_speedx[MAXPTS];
PyFloatObject *pf_speedy[MAXPTS];
PyFloatObject *pf_speedz[MAXPTS];

PyObject *pt_coord[MAXPTS];
PyObject *pt_speed[MAXPTS];

#define PTC(i) pt_coord[i]
#define PTS(i) pt_speed[i]

#define X(i) (coordx[i])
#define Y(i) (coordy[i])
#define Z(i) (coordz[i])

#define PX(i) (PFX(i)->ob_fval)
#define PY(i) (PFY(i)->ob_fval)
#define PZ(i) (PFZ(i)->ob_fval)
#define PFX(i) pf_coordx[i]
#define PFY(i) pf_coordy[i]
#define PFZ(i) pf_coordz[i]
     
#define SX(i) (speedx[i])
#define SY(i) (speedy[i])
#define SZ(i) (speedz[i])

#define PSX(i) (PFSX(i)->ob_fval)
#define PSY(i) (PFSY(i)->ob_fval)
#define PSZ(i) (PFSZ(i)->ob_fval)
#define PFSX(i) pf_speedx[i]
#define PFSY(i) pf_speedy[i]
#define PFSZ(i) pf_speedz[i]

#define FX(i) forcex[i]
#define FY(i) forcey[i]
#define FZ(i) forcez[i]

#define CONDADD(idx, len, arr) do{ switch (len-idx) { default: (arr)[2] += (arr)[3]; case 3: (arr)[1] += (arr)[2]; case 2: (arr)[0] += (arr)[1]; case 1:; } } while(0);

int mmx_control_flags = 0x5f80; // 0x5fc0; // set FTZ and DAZ
                                    
int old_len;

float tmpshow[4] __attribute__ ((aligned(32)));
#define SHOW(r,fmt,args ...) do { __asm__("lea tmpshow,%%edx \n\t movaps " r ",(%%edx) \n\t":::"edx"); \
          printf("[%+.3f %+.3f %+.3f %+.3f] " fmt "\n" , tmpshow[0],tmpshow[1],tmpshow[2],tmpshow[3], ## args );} while(0)
#define LOG(x ...)
//#define LOG(x ...) printf(x)
         

#define EDGEW1 16
#define EDGELW1 4
#define EDGEW2 16
#define EDGELW2 4
#define EDGEW3 (8*sizeof(int))
#define EDGELW3 5

struct edge_part2 {
    int val[EDGEW2];
};

struct edge_part1 {
    struct edge_part2 *next[EDGEW1];
};


struct edge_part1 *edges[MAXPTS];

// we always have i <= j
int check_edge(int i, int j)
{
        struct edge_part1 *p1;
        struct edge_part2 *p2;
        int n1,n2,n3;

        p1 = edges[i];
        if (!p1) return 0;

        n1 = (j >> (2*EDGELW1)) & (EDGEW1-1);
        p2 = p1->next[n1];
        if (!p2) return 0;
        
        n2 = (j >> EDGELW2) & (EDGEW2-1);
        n3 = j & (EDGEW3-1);

        return (p2->val[n2] & (1<<n3)) != 0;
}

void do_add_edge(int i, int j)
{
        struct edge_part1 *p1;
        struct edge_part2 *p2;
        int n1,n2,n3;

        n1 = (j >> (2*EDGELW1)) & (EDGEW1-1);
        n2 = (j >> EDGELW2) & (EDGEW2-1);
        n3 = j & (EDGEW3-1);
        
        p1 = edges[i];
        if (!p1) {
                p1 = edges[i] = malloc(sizeof(struct edge_part1));
                bzero(p1, sizeof(struct edge_part1));
        }
        p2 = p1->next[n1];
        if (!p2) {
                p2 = p1->next[n1] = malloc(sizeof(struct edge_part2));
                bzero(p2, sizeof(struct edge_part2));
        }
        p2->val[n2] |= 1<<n3;
}


void add_edge(int i, int j)
{
      if (i < j) { do_add_edge(i,j); } // check is always called with i < j
      else { do_add_edge(j,i); }
}


// SSE prepa

float attract[4] __attribute__ ((aligned(32)));
float repuls[4] __attribute__ ((aligned(32)));
float min_dist[4] __attribute__ ((aligned(32))) = { MIN_DIST, MIN_DIST, MIN_DIST, MIN_DIST };
float friction[4] __attribute__ ((aligned(32))) = { FRICTION, FRICTION, FRICTION };
float zero[4] __attribute__ ((aligned(32))) = { 0.0, 0.0, 0.0, 0.0 };
float half[4] __attribute__ ((aligned(32))) = { 0.5, 0.5, 0.5, 0.5 };
float tmpsse1[4] __attribute__ ((aligned(32)));
float tmpsse2[4] __attribute__ ((aligned(32)));
float tmpsse3[4] __attribute__ ((aligned(32)));


float update(PyObject *the_world, PyObject *pos, PyObject *spd, int changed, float dt, float ATTRACT, float REPULS)
{
	PyObject *item;
	int i,j,len;
        float dA;

	len = PyList_GET_SIZE(pos);
	if (len > MAXPTS) len = MAXPTS;

        repuls[0] = repuls[1] = repuls[2] = repuls[3] = REPULS;

        LOG("=== changed=%i old_len=%04i len=%04i attract=%+.f repuls=%+.2f\n",
               changed, old_len, len, ATTRACT, REPULS);

        if (changed  || (old_len != len)) {
                for (i=old_len; i<len; i++) {
                        PFX(i)  = (PyFloatObject *)PyFloat_FromDouble(0.0);
                        PFY(i)  = (PyFloatObject *)PyFloat_FromDouble(0.0);
                        PFZ(i)  = (PyFloatObject *)PyFloat_FromDouble(0.0);
                        PTC(i) = item = PyTuple_New(3);
                        PyTuple_SET_ITEM(item, 0, (PyObject *)PFX(i));
                        PyTuple_SET_ITEM(item, 1, (PyObject *)PFY(i));
                        PyTuple_SET_ITEM(item, 2, (PyObject *)PFZ(i));
                        
                        PFSX(i) = (PyFloatObject *)PyFloat_FromDouble(0.0);
                        PFSY(i) = (PyFloatObject *)PyFloat_FromDouble(0.0);
                        PFSZ(i) = (PyFloatObject *)PyFloat_FromDouble(0.0);
                        PTS(i) = item = PyTuple_New(3);
                        PyTuple_SET_ITEM(item, 0, (PyObject *)PFSX(i));
                        PyTuple_SET_ITEM(item, 1, (PyObject *)PFSY(i));
                        PyTuple_SET_ITEM(item, 2, (PyObject *)PFSZ(i));
                }
                old_len = len;
	    	for (i=0; i<len; i++) {
                        item = PyList_GET_ITEM(pos, i);
                        if (PTC(i) != item) {
                                X(i) = ((PyFloatObject *)PyTuple_GET_ITEM(item, 0))->ob_fval;
                                Y(i) = ((PyFloatObject *)PyTuple_GET_ITEM(item, 1))->ob_fval;
                                Z(i) = ((PyFloatObject *)PyTuple_GET_ITEM(item, 2))->ob_fval;
                                Py_INCREF(PTC(i));
                                PyList_SET_ITEM(pos, i, PTC(i));
                        }
                        
                        item = PyList_GET_ITEM(spd, i);
                        if (PTS(i) != item) {
                                SX(i) = ((PyFloatObject *)PyTuple_GET_ITEM(item, 0))->ob_fval;
                                SY(i) = ((PyFloatObject *)PyTuple_GET_ITEM(item, 1))->ob_fval;
                                SZ(i) = ((PyFloatObject *)PyTuple_GET_ITEM(item, 2))->ob_fval;
                                Py_INCREF(PTS(i));
                                PyList_SET_ITEM(spd, i, PTS(i));
                        }

	    	}
        }
        


        asm("ldmxcsr %0 \n\t" :: "m"(mmx_control_flags)); // set DAZ and MTZ into MMX control reg

        asm("movaps %0, %%xmm0 \n\t" ::"m"(friction[0]));
	for (i=0; i<len; i+=4) {
                asm("movups %0, %%xmm1"::"m"(SX(i)));
                asm("movups %0, %%xmm2"::"m"(SY(i)));
                asm("movups %0, %%xmm3"::"m"(SZ(i)));
                asm("mulps %%xmm0, %%xmm1 \n\t"
                    "mulps %%xmm0, %%xmm2 \n\t"
                    "mulps %%xmm0, %%xmm3 \n\t":);
                asm("movups %%xmm1,%0"::"m"(FX(i)));
                asm("movups %%xmm2,%0"::"m"(FY(i)));
                asm("movups %%xmm3,%0"::"m"(FZ(i)));
        }

	for (i=0; i<len; i++) {

                asm("movss %0,%%xmm0\n\t"
                    "movss %1,%%xmm1\n\t"
                    "movss %2,%%xmm2\n\t"
                    "shufps $0, %%xmm0, %%xmm0 \n\t"
                    "shufps $0, %%xmm1, %%xmm1 \n\t"
                    "shufps $0, %%xmm2, %%xmm2 \n\t"
                    ::"m"(X(i)),"m"(Y(i)),"m"(Z(i)));

		for (j=i+1; j<len; j+=4) {

                        asm("movups %0, %%xmm3 \n\t"
                            "movups %1, %%xmm4 \n\t"
                            "movups %2, %%xmm5 \n\t"
                            "subps %%xmm0, %%xmm3 \n\t" // xmm3=dX
                            "subps %%xmm1, %%xmm4 \n\t" // xmm4=dY
                            "subps %%xmm2, %%xmm5 \n\t" // xmm5=dZ
                            ::"m"(X(j)),"m"(Y(j)),"m"(Z(j)));

                        attract[0] = ATTRACT*check_edge(i,j);
                        attract[1] = ATTRACT*check_edge(i,j+1);
                        attract[2] = ATTRACT*check_edge(i,j+2);
                        attract[3] = ATTRACT*check_edge(i,j+3);

                        LOG("attract %2i %2i: %5i %5i %5i %5i\n", i,j,attract[0],attract[1],attract[2],attract[3]);

			if (attract[0] || attract[1] || attract[2] || attract[3]) {
                                asm("movaps %0, %%xmm6 \n\t"
                                    "mulps %%xmm3, %%xmm6 \n\t"
                                    "movups %2, %%xmm7 \n\t"
                                    "movaps %%xmm6, %1 \n\t"
                                    "subps %%xmm6, %%xmm7 \n\t"
                                    "movups %%xmm7, %2 \n\t"::"m"(attract[0]), "m"(tmpsse1[0]), "m"(FX(j)));
                                LOG("attract X: %+.3f %+.3f %+.3f %+.3f\n",
                                       tmpsse1[0], tmpsse1[1], tmpsse1[2], tmpsse1[3]);
                                       
                                asm("movaps %0, %%xmm6 \n\t"
                                    "mulps %%xmm4, %%xmm6 \n\t"
                                    "movups %2, %%xmm7 \n\t"
                                    "movaps %%xmm6, %1 \n\t"
                                    "subps %%xmm6, %%xmm7 \n\t"
                                    "movups %%xmm7, %2 \n\t"::"m"(attract[0]), "m"(tmpsse2[0]), "m"(FY(j)));
                                LOG("attract Y: %+.3f %+.3f %+.3f %+.3f\n",
                                       tmpsse2[0], tmpsse2[1], tmpsse2[2], tmpsse2[3]);    

                                asm("movaps %0, %%xmm6 \n\t"
                                    "mulps %%xmm5, %%xmm6 \n\t"
                                    "movups %2, %%xmm7 \n\t"
                                    "movaps %%xmm6, %1 \n\t"
                                    "subps %%xmm6, %%xmm7 \n\t"
                                    "movups %%xmm7, %2 \n\t"::"m"(attract[0]), "m"(tmpsse3[0]), "m"(FZ(j)));
                                LOG("attract Z: %+.3f %+.3f %+.3f %+.3f\n",
                                       tmpsse3[0], tmpsse3[1], tmpsse3[2], tmpsse3[3]);


                                if (attract[0]) { FX(i) += tmpsse1[0]; FY(i) += tmpsse2[0]; FZ(i) += tmpsse3[0];}
                                if (attract[1]) { FX(i) += tmpsse1[1]; FY(i) += tmpsse2[1]; FZ(i) += tmpsse3[1];}
                                if (attract[2]) { FX(i) += tmpsse1[2]; FY(i) += tmpsse2[2]; FZ(i) += tmpsse3[2];}
                                if (attract[3]) { FX(i) += tmpsse1[3]; FY(i) += tmpsse2[3]; FZ(i) += tmpsse3[3];}

			}


                        asm("movaps %%xmm3, %%xmm7 \n\t"
                            "movaps %%xmm4, %%xmm6 \n\t"
                            "mulps %%xmm3, %%xmm7 \n\t"
                            "mulps %%xmm4, %%xmm6 \n\t"
                            "addps %%xmm6, %%xmm7 \n\t"
                            "movaps %%xmm5, %%xmm6 \n\t"
                            "mulps %%xmm5, %%xmm6 \n\t"
                            "addps %%xmm6, %%xmm7 \n\t" // xmm7 = D = dX*dX+dY*dY+dZ*dZ
                            "mulps %%xmm7, %%xmm7 \n\t"
                            "maxps %0, %%xmm7 \n\t"     // xmm7 = max(MIN_DIST, D^2)
                            ::"m"(min_dist[0])); 

                        asm("movaps %0, %%xmm6 \n\t"  // xmm6 = repuls
                            "divps %%xmm7, %%xmm3 \n\t"
                            "divps %%xmm7, %%xmm4 \n\t"
                            "divps %%xmm7, %%xmm5 \n\t"
                            "mulps %%xmm6, %%xmm3 \n\t"
                            "mulps %%xmm6, %%xmm4 \n\t"
                            "mulps %%xmm6, %%xmm5 \n\t" // xmm3,4,5 = dfx,dfy,dfz
                            :: "m"(repuls[0]));

                        asm("movaps %%xmm3, %0\n\t"
                            "movaps %%xmm4, %1\n\t"
                            "movaps %%xmm5, %2\n\t"
                            ::"m"(tmpsse1[0]),"m"(tmpsse2[0]),"m"(tmpsse3[0]));
                            
                        asm("movups %0, %%xmm6 \n\t"
                            "movups %1, %%xmm7 \n\t"
                            "addps %%xmm3, %%xmm6 \n\t"
                            "addps %%xmm4, %%xmm7 \n\t"
                            "movups %2, %%xmm3 \n\t"
                            "movups %%xmm6, %0 \n\t"
                            "addps %%xmm5, %%xmm3 \n\t"
                            "movups %%xmm7, %1 \n\t"
                            "movups %%xmm3, %2 \n\t"
                            ::"m"(FX(j)), "m"(FY(j)), "m"(FZ(j)));
                            
                        CONDADD(j,len,tmpsse1);
                        FX(i) -= tmpsse1[0];
                        CONDADD(j,len,tmpsse2);
                        FY(i) -= tmpsse2[0];
                        CONDADD(j,len,tmpsse3);
                        FZ(i) -= tmpsse3[0];
		}
	}

        dt *= ACCEL;
        asm("movaps %0, %%xmm7 \n\t"    // xmm7 = dA = 0
            "movaps %1, %%xmm6 \n\t"    // xmm6 = 0.5
            "movss %2,%%xmm0 \n\t"
            "shufps $0, %%xmm0, %%xmm0 \n\t" // xmm1=(dt,dt,dt,dt)
                    ::"m"(zero[0]),"m"(half[0]), "m"(dt));
        
	for (i=0; i<len; i+=4) {

                asm("movups %0,%%xmm1 \n\t" // xmm1=FX
                    "movups %1,%%xmm2 \n\t" // xmm2=SX
                    "movups %2,%%xmm3 \n\t" // xmm3=X
                    "mulps %%xmm0, %%xmm1 \n\t" // xmm1=F*dt
                    "addps %%xmm1, %%xmm2 \n\t" // xmm2=S+dS=S+Fdt
                    "mulps %%xmm6, %%xmm1 \n\t"
                    "addps %%xmm2, %%xmm1 \n\t"
                    "mulps %%xmm0, %%xmm1 \n\t" // xmm1 = (S+0.5*F*dt)*dt = dX
                    "addps %%xmm1, %%xmm3 \n\t"
                    "movups %%xmm3, %2 \n\t"
                    "mulps %%xmm3,%%xmm3 \n\t"
                    "movups %%xmm2, %1 \n\t"
                    "addps %%xmm3, %%xmm7 \n\t" //xmm7 = dA += dX^2
                    ::"m"(FX(i)),"m"(SX(i)),"m"(X(i)));

                asm("movups %0,%%xmm1 \n\t" // xmm1=FY
                    "movups %1,%%xmm2 \n\t" // xmm2=SY
                    "movups %2,%%xmm3 \n\t" // xmm3=Y
                    "mulps %%xmm0, %%xmm1 \n\t" // xmm1=F*dt
                    "addps %%xmm1, %%xmm2 \n\t" // xmm2=S+dS=S+Fdt
                    "mulps %%xmm6, %%xmm1 \n\t"
                    "addps %%xmm2, %%xmm1 \n\t"
                    "mulps %%xmm0, %%xmm1 \n\t" // xmm1 = (S+0.5*F*dt)*dt = dY
                    "addps %%xmm1, %%xmm3 \n\t"
                    "movups %%xmm3, %2 \n\t"
                    "mulps %%xmm3,%%xmm3 \n\t"
                    "movups %%xmm2, %1 \n\t"
                    "addps %%xmm3, %%xmm7 \n\t" //xmm7 = dA += dX^2
                    ::"m"(FY(i)),"m"(SY(i)),"m"(Y(i)));

                asm("movups %0,%%xmm1 \n\t" // xmm1=FZ
                    "movups %1,%%xmm2 \n\t" // xmm2=SZ
                    "movups %2,%%xmm3 \n\t" // xmm3=Z
                    "mulps %%xmm0, %%xmm1 \n\t" // xmm1=F*dt
                    "addps %%xmm1, %%xmm2 \n\t" // xmm2=S+dS=S+Fdt
                    "mulps %%xmm6, %%xmm1 \n\t"
                    "addps %%xmm2, %%xmm1 \n\t"
                    "mulps %%xmm0, %%xmm1 \n\t" // xmm1 = (S+0.5*F*dt)*dt = dZ
                    "addps %%xmm1, %%xmm3 \n\t"
                    "movups %%xmm3, %2 \n\t"
                    "mulps %%xmm3,%%xmm3 \n\t"
                    "movups %%xmm2, %1 \n\t"
                    "addps %%xmm3, %%xmm7 \n\t" //xmm7 = dA += dX^2
                    ::"m"(FZ(i)),"m"(SZ(i)),"m"(Z(i)));
                    
                }
        asm("movaps %%xmm7, %0"::"m"(tmpsse1[0]));
        dA = tmpsse1[0]+tmpsse1[1]+tmpsse1[2]+tmpsse1[3];

        for (i=0; i<len; i++) {
            PX(i) = X(i);
            PY(i) = Y(i);
            PZ(i) = Z(i);
        }

	return dA/(len+0.1);
}
		
""")
    C_ENGINE=1
except ImportError:
    log.warning("PyInline module not found! Using fallback python physics engine, which is 25 to 100 times slower")
    physics_engine = python_physics_engine
    C_ENGINE=0


class Edge(cylinder):
    def __init__(self, o1, o2, *args, **kargs):
        param={"radius":0.2, "color":(0.7,0.7,1)}
        param.update(kargs)
        self.o1=o1
        self.o2=o2
        cylinder.__init__(self, *args, **param)
        self.update()
        self.pickable = 0
    def update(self):
        self.pos = self.o1.pos
        self.axis = self.o2.pos-self.o1.pos

def randvect():
    return vector(random.random(),random.random(),random.random())

def randcol():
    v = randvect()
    s = v[0]+v[1]+v[2]
    if s < 1:
        v *= 1/(0.001+s)        
    return v

class Node(sphere):
    def __init__(self, name, *args, **kargs):
        self.param = {"radius":1, "pos": randvect(), "color": randcol() }
        self.param.update(kargs)

        self.name = name
        self.edges = {}
        self.pickable = 1
        self.PV=vector()
        self.V=vector()

        sphere.__init__(self, *args, **self.param)

        self.label=label(text=name, pos=self.pos, space=self.radius, xoffset=10, yoffset=20, visible=0)
        self.update_label()


    def update_label(self):
        text = [self.name]
        for k,v in self.param.iteritems():
            if k in ["radius", "pos", "color"]:
                continue
            text.append("%s = %s" % (k, v))
        self.label.text = "\n".join(text)

    def update_param(self, **kargs):
        self.param.update(kargs)
        self.update_label()
    def action(self):
        self.label.pos=self.pos
        self.label.visible ^= 1
    def update(self, dt):
        self.label.pos = self.pos
    def get_state(self):
        for k in self.param:
            if hasattr(self, k):
                self.param[k] = getattr(self,k)
        return self.param
    def dump(self):
        return repr( (self.name, self.get_state(), map(lambda x:x.name,self.edges.keys())) )
    @classmethod
    def from_string(cls, s):
        l = eval(s)
        name,state,edges = l[:3]
        o = cls(name, **state)
        o._tmp_edges = edges
        return o

povexport.legal[Node] = povexport.legal[sphere]
povexport.legal[Edge] = povexport.legal[cylinder]


class World:
    def __init__(self, friction=-0.5, attraction=1, repulsion=256):
        self.friction = friction
        self.attraction = attraction
        self.repulsion = repulsion
        self.world={}
        self.lst = []
        self.world_len = 0
        self.edges = []
        self.coords = []
        self.speeds = []
        self.cinematic_change = 1
    def register(self, o):
        self.world[o.name] = o
        self.lst.append(o)
        o.num = self.world_len
        self.coords.append(o.pos.astuple())
        self.speeds.append((0.0,0.0,0.0))
        self.world_len += 1
    def add_edge(self, o1, o2):
        self.edges.append(Edge(o1,o2))
        physics_engine.add_edge(o1.num, o2.num)
    def __getitem__(self, item):
        return self.world[item]
    def __contains__(self,item):
        return item in self.world
    def get(self, *args):
        return self.world.get(*args)
    def dump_to_file(self, f):
        for o in self.lst:
            f.write(o.dump())
            f.write("\n")
        f.close()
    def load_from_file(self, f):
        while 1:
            l = f.readline()
            if not l:
                break
            o = Node.from_string(l)
            self.register(o)
        for o in self.lst:
            for e in o._tmp_edges:
                if e not in self.world:
                    print >>sys.stderr, "%s: edge to [%s] not found" % (o.name,e)
                    continue
                o2 = self.world[e]
                o.edges[o2] = None
                if o in o2.edges:
                    self.add_edge(o,o2)
            del(o._tmp_edges)
    def update_edges(self):
        for e in self.edges:
            e.update()        
    def update(self, dt):
        physics_engine.update(self, self.coords, self.speeds, self.cinematic_change, dt, self.attraction, self.repulsion)
        self.cinematic_change = 0
        if C_ENGINE:
            for i in xrange(self.world_len):
                self.lst[i].pos = self.coords[i]
                self.lst[i].update(dt)
        self.update_edges()



def mkcol():
    def norm(x):
        if x > 1 or x <= -1:
            x = 1
        elif x < 0:
            x += 1
        return x
    return (random.random(), random.random(), random.random())

def ip2num(ip):
    a,b,c,d = struct.unpack("!lLlL",socket.inet_pton(socket.AF_INET6, ip))
    return (a<<32L|b), (c<<32L|d)


class DBquery:
    def __init__(self, db):
        if db is not None:
            log.info("Opening net DB [%s]" % db)
            self.con = sqlite.connect(db)
            self.cur = self.con.cursor()
            self.color = {None:(0.7,0.7,0.7)}
        else:
            self.ip2attr = lambda x: {"color":randcol()}

    def ip2attr(self, ip):
        try:
            ipnum = ip2num(ip)
        except:
            return { "color":randcol() }
        log.info("Query DB for IP [%s]" % ip)
        self.cur.execute("""select netname,country.name,country.a2
                            from inet6num
                                 left outer join inetinfo on inetinfo_id=inetinfo.id
                                 left outer join country on country_id=country.id
                            where prefix_hi = ? & mask_hi and
                                  prefix_lo = ? & mask_lo
                            order by length desc""", ipnum)
        res = self.cur.fetchone()
        if res is None:
            log.error("IP [%s] not found" % ip)
            res = "??","??",None
        netname,country,a2 = res
        if a2 not in self.color:
            self.color[a2] = mkcol()
        return {"netname":netname, "country":country, "color":self.color[a2]}

class Cinematic:
    static=0
    dynamic=1
    dyn_thread=2

frames = 0

def mainloop(input=[sys.stdin], startfile=None, savefile=None, stereo=None, cinematic=Cinematic.dynamic, POVdump=None,dbname=None):
    global frames
    scene.title="RealTime Graph 3D"
    scene.exit=0
    if stereo is not None:
        scene.stereo = stereo
        scene.stereodepth = 2
#        scene.width=100
#        scene.height=50
    scene.visible=1

    dbq = DBquery(dbname)
    picked = None
    the_world = World()
    if startfile is not None:
        the_world.load_from_file(open(startfile))
    if savefile is not None:
        atexit.register(the_world.dump_to_file, open(savefile,"w"))

    def cinematic_thread(POVdump=None):
        log.info("Cinematic thread started")
        global frames
        while 1:
            try:
                frames += 1
                rate(ACCEL/DT)
                the_world.update(DT)
                if POVdump is not None:
                    fname = POVdump % frames
                    povexport.export(filename=fname, display=scene, include_list=['colors.inc'])
            except:
                log.exception("FRAME=%i: Catched exception in the cinematic thread!" % frames)


    if cinematic == Cinematic.dyn_thread:
        thread.start_new_thread(cinematic_thread, (POVdump,))

    starttime=time.time()
    try:
        while 1:
            if cinematic == Cinematic.dynamic:
                frames += 1
                rate(ACCEL/DT)
                i,_,_ = select(input,[],[],0)
            else:
                i,_,_ = select(input,[],[],0.05)
            for fd in i:
                try:
                    for cmd in os.read(fd.fileno(),1500).strip().splitlines():
                        try:
                            log.debug("Got [%s]" % cmd)
                            if cmd.startswith("dump"):
                                if " " in cmd:
                                    fname = cmd.split()[1]
                                else:
                                    fname = "/dev/stdout"
                                the_world.dump_to_file(open(fname,"w"))
                                log.info("Dumped world to [%s]" % fname)
                            elif cmd.startswith("dot"):
                                if " " in cmd:
                                    fname = cmd.split()[1]
                                else:
                                    fname = "/dev/stdout"
                                f = open(fname,"w")
                                f.write('graph "rtgraph" { \n')
                                for e in the_world.edges:
                                    f.write('\t "%s" -- "%s";\n' % (e.o1.name, e.o2.name))
                                f.write('}\n')
                                f.close()
                                log.info("Wrote dot file to [%s]" % fname)
                            elif cmd.startswith("attract"):
                                the_world.attraction = float(cmd.split()[1])
                            elif cmd.startswith("repuls"):
                                the_world.repulsion = float(cmd.split()[1])
                            elif cmd.startswith("stereodepth"):
                                scene.stereodepth = int(cmd.split()[1])
                            elif cmd.startswith("tog"):
                                name = cmd.split()[1]
                                n = the_world.get(name)
                                if n:
                                    if hasattr(n, "action"):
                                        n.action()
                                else:
                                    log.info("[%r] not found" % name)
                            elif cmd.startswith("exec"):
                                cmd = cmd.split(" ",1)[1]
                                log.info("--- Begin exec [%s]" % cmd) 
                                res=eval(cmd)
                                log.info(res)
                                log.info("--- End exec [%s]" % cmd) 
                            elif cmd.startswith("quit"):
                                raise KeyboardInterrupt
                                
                            elif cmd.startswith("check"):
                                for o in the_world.lst:
                                    for o2 in o.edges:
                                        if o not in o2.edges:
                                            print "%s ==> %s \t %s ==> %s"% (o.name,o2.name, id(o),id(o2))
                                log.info("Checked %i nodes, %i edges." % (len(the_world.lst),len(the_world.edges)))
        
                            elif cmd:
                                nam1,nam2 = map(string.strip, cmd.split())
        
                                if nam1 == nam2:
                                    raise Exception("Nodes are identical")
            
                                n1 = the_world.get(nam1)
                                n2 = the_world.get(nam2)
                                if n1 is None:
                                    n2,n1=n1,n2
                                    nam1,nam2=nam2,nam1
            
                                if n1 is None:
                                    attr=dbq.ip2attr(nam1)
                                    n1 = Node(nam1,**attr)
                                    the_world.register(n1)
                                if n2 is None:
                                    attr=dbq.ip2attr(nam2)
                                    n2 = Node(nam2, pos=n1.pos+randvect(), **attr)
                                    the_world.register(n2)
            
                                if n1 in n2.edges:
                                    raise Exception("Edge %s-%s alread exists" % (nam1,nam2))
                
                                n1.edges[n2] = None
                                n2.edges[n1] = None
                                the_world.add_edge(n1,n2)
                                the_world.cinematic_change = 1
                                
                                log.info("Ok for [%s]-[%s]" % (nam1,nam2))
                        except KeyboardInterrupt:
                            raise
                        except:
                            log.exception("")
                                
                except KeyboardInterrupt:
                    raise
                except:
                    log.exception("")
            if scene.kb.keys:
                k = scene.kb.getkey()
                if k == "Q":
                    raise KeyboardInterrupt
            if scene.mouse.events:
                ev = scene.mouse.getevent()
                if ev.release == "left":
                    if picked:
                        the_world.speeds[picked.num]=((scene.mouse.pos-picked.PV)/DT).astuple()
                        the_world.cinematic_change = 1
                        picked = None
                    else:
                        o = ev.pick
                        if o and hasattr(o, "action"):
                            o.action()
                if ev.drag == "left":
                    if ev.pick and ev.pick.pickable:
                        picked = ev.pick

            if cinematic == Cinematic.dynamic:
                the_world.update(DT)
            if cinematic == Cinematic.static:
                the_world.update_edges()
    
            if picked:
                picked.PV = picked.pos
                picked.pos=scene.mouse.pos
                picked.V = vector()
                the_world.coords[picked.num]=picked.pos.astuple()
                the_world.speeds[picked.num]=(0.0,0.0,0.0)
                the_world.cinematic_change = 1

    except KeyboardInterrupt:
        log.info("Interrupted by user")
    elapsed = time.time()-starttime
    log.info("Frame rate: wanted=%.3ffps actual=%.3ffps" % (ACCEL/DT,frames/elapsed))
        

def usage():
    print >>sys.stderr, """rtgraph3D [-l [ip:]port [-t]] [-p] [-w savefile] [-r startfile]
    -l [ip:]port    listen socket
    -t              TCP mode
    -p              profile application
    -P <fname_tmpl> export to POV, filename template (ex: /tmp/foo/bar-%i.pov)
    -d 
    -c {s|d|t}      Choobe between static, dynamic and threaded dynamic mode
    -w <savefile>   save world in this file
    -r <startfile>  read world from this file
    -s <stereo>     stereo mode (active, passive, crosseyed, redcyan, redblue, yellowblue)"""
    sys.exit(0)

if __name__ == "__main__":
    INPUT = [sys.stdin]
    PROFILE=0
    NETADDR=None
    MODE=socket.SOCK_DGRAM
    STARTFILE=None
    SAVEFILE=None
    STEREO=None
    CINEMATIC=Cinematic.dynamic
    POVDUMP=None
    DBNAME=None
    try:
        opts = getopt.getopt(sys.argv[1:],"hl:ptr:w:s:c:P:d:")

        for opt,optarg in opts[0]:
            if opt == "-h":
                usage()
            elif opt == "-p":
                PROFILE = 1
            elif opt == "-t":
                MODE = socket.SOCK_STREAM
            elif opt == "-r":
                STARTFILE = optarg
            elif opt == "-s":
                STEREO = optarg
            elif opt == "-w":
                SAVEFILE = optarg
            elif opt == "-P":
                POVDUMP = optarg
            elif opt == "-d":
                DBNAME = optarg
            elif opt == "-c":
                if optarg.startswith("s"): # static
                    CINEMATIC = Cinematic.static
                elif optarg.startswith("d"): # dynamic
                    CINEMATIC = Cinematic.dynamic
                elif optarg.startswith("t"): #thread
                    CINEMATIC = Cinematic.dyn_thread
            elif opt == "-l":
                try:
                   if ":" in optarg:
                       ip,port = optarg.split(":")
                   else:
                       ip,port = "",optarg
                   port = int(port)
                   NETADDR=ip,port
                except:
                    raise getopt.GetoptError("Failed to parse [%s]" % optarg)
    except getopt.GetoptError,msg:
        log.exception(msg)
        sys.exit(-1)

    if NETADDR:
        s = socket.socket(socket.AF_INET, MODE)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((ip,port))
        if MODE == socket.SOCK_STREAM:
            s.listen(1)
            t,f = s.accept()
            log.info("Received TCP connection from %s:%i" % f)
            s.close()
            s = t
        INPUT.append(s)
    

    if PROFILE:
        log.info("Profiling mode")
        import profile
        profile.runctx('mainloop(INPUT)',globals=globals(), locals=locals())
        log.info("End of profiling")
        
    else:
        mainloop(INPUT,startfile=STARTFILE, savefile=SAVEFILE, stereo=STEREO, cinematic=CINEMATIC,
                 POVdump=POVDUMP, dbname=DBNAME)




